import hashlib
import json
import multiprocessing
import os
from pathlib import Path
from typing import Any, Dict


class FluxCache:
    _lock = multiprocessing.Lock()
    _manager = multiprocessing.Manager()

    def __init__(self, cache_file: str | None = None):
        """
        Initialize the API cache manager

        Args:
            cache_file: path to the cache file
        """
        if cache_file is None:
            cache_file = os.getenv("CACHE_FILE", "cache.jsonl")
        self.cache_file = Path(cache_file)
        self.cache_data: Dict[str, Any] = self._manager.dict()
        self._load_cache()

    def _load_cache(self):
        """
        Load the cache file
        """
        if self.cache_file.exists():
            with open(self.cache_file, 'r', encoding='utf-8') as f:
                cache_data = {item["id"]: item for item in (json.loads(line) for line in f)}
            print(f"Found {len(cache_data)} cached samples")
        else:
            Path(self.cache_file).parent.mkdir(parents=True, exist_ok=True)
            cache_data = {}
            print(f"Initialized empty cache")

        self.cache_data.update(cache_data)

    def hash(self, sample: Dict) -> str:
        """
        Generate a unique hash ID for the input sample

        Args:
            sample: request sample
        Returns:
            generated hash ID
        """
        sorted_str = json.dumps(sample, sort_keys=True)
        return hashlib.md5(sorted_str.encode('utf-8')).hexdigest()

    def save_to_cache(self, sample: Dict, response: Dict, save_request: bool = False):
        """
        Add the API response to the cache
        
        Args:
            sample: original request sample
            response: API response result
            save_request: whether to save the request sample
        """
        sample_id = self.hash(sample)
        if save_request:
            cache_entry = {'id': sample_id, 'request': sample, 'response': response}
        else:
            cache_entry = {'id': sample_id, 'response': response}

        with self._lock:
            with open(self.cache_file, 'a', encoding='utf-8') as f:
                f.write(json.dumps(cache_entry, ensure_ascii=False) + '\n')
            self.cache_data[sample_id] = cache_entry

    def is_cached(self, sample: Dict) -> bool:
        """
        Check if the data is already cached

        Args:
            sample: request sample
        Returns:
            True if the sample is cached, False otherwise
        """
        sample_id = self.hash(sample)
        return sample_id in self.cache_data

    def collect_result(self, sample: Dict) -> Dict | None:
        """
        Get the cached response corresponding to the sample.

        Args:
            sample: request sample
        Returns:
            cached response associated with the sample
        """
        sample_id = self.hash(sample)
        if sample_id not in self.cache_data:
            return None
        return self.cache_data[sample_id]['response']



import os
from typing import Any, Dict, List, Optional

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion
from openai.types.completion import Completion

from fluxllm.clients.base import BaseClient


class FluxOpenAIChat(BaseClient):
    """
    A client that can handle multiple requests concurrently for OpenAI chat completions.
    """

    SUPPORT_ARGS = {
        "model",
        "messages",
        "frequency_penalty",
        "logit_bias",
        "logprobs",
        "top_logprobs",
        "max_tokens",
        "n",
        "presence_penalty",
        "response_format",
        "seed",
        "stop",
        "stream",
        "temperature",
        "top_p",
        "tools",
        "tool_choice",
        "user",
        "function_call",
        "functions",
        "tenant",
        "max_completion_tokens",
    }

    progress_prompt: str = "Generating..."

    def __init__(
        self,
        cache_file: str | None = None,
        max_retries: int | None = None,
        base_url: str | None = None,
        api_key: str | None = None,
        max_qps: Optional[float] = None,
        max_qpm: float = 100,
    ):
        """
        Initialize the client.

        Args:
            - cache_file: the file to cache the results.
            - max_retries: the maximum number of retries to make, defaults to `None`.
            - base_url: OpenAI Compatible API base URL.
            - api_key: OpenAI API key.
            - max_qps: maximum queries per second (rate limit), defaults to `None` (falls back to max_qpm).
            - max_qpm: maximum queries per minute (rate limit), defaults to `100`.
        """

        super().__init__(
            cache_file=cache_file, 
            max_retries=max_retries, 
            max_qps=max_qps,
            max_qpm=max_qpm
        )

        # initialize the client
        self.client = AsyncOpenAI(
            base_url=base_url or os.getenv("OPENAI_API_BASE"),
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
        )

    async def make_request_async(self, request: Dict[str, Any], **kwargs) -> Dict | None:
        """
        Make a single request to the OpenAI API and cache the response.

        Args:
            request: the request to make
            **kwargs: Additional arguments that will be passed to the OpenAI API call if they are in SUPPORT_ARGS

        Returns:
            bool: True if the request was successful and cached, False if it failed
        """

        try:

            assert "messages" in request, "`messages` are required in the request"
            assert "model" in request, "`model` is required in the request"

            gen_kwargs = {}
            for key in self.SUPPORT_ARGS:
                if key in request:
                    gen_kwargs[key] = request[key]
                else:
                    if key in kwargs and kwargs[key] is not None:
                        gen_kwargs[key] = kwargs[key]
            # add timeout if it is specified
            if "timeout" in kwargs:
                gen_kwargs["timeout"] = kwargs["timeout"]
            # request for the response
            response: ChatCompletion = await self.client.chat.completions.create(**gen_kwargs)
            response_dict = response.model_dump()
            return response_dict

        except Exception as e:
            print(f"{type(e).__name__}: {e}")
            return None

    def collect_responses(self, requests: List[Dict]) -> List[ChatCompletion]:
        """
        Get the cached responses for a list of requests and convert them to `ChatCompletion` objects.

        Args:
            requests: list of requests to get responses for
        Returns:
            list of ChatCompletion objects containing the cached responses
        """

        # First get the cached responses as dictionaries
        cached_responses = [self.cache.collect_result(request) for request in requests]

        # Then convert dictionaries to ChatCompletion objects
        chat_completions = []
        for cached_response in cached_responses:
            if cached_response is None:
                chat_completion = None
            else:
                chat_completion = ChatCompletion.model_validate(cached_response)
            chat_completions.append(chat_completion)

        return chat_completions


class FluxOpenAICompletion(BaseClient):
    """
    A client that can handle multiple requests concurrently for OpenAI text completions.
    """

    SUPPORT_ARGS = {
        "model",
        "prompt",
        "best_of",
        "echo",
        "frequency_penalty",
        "logit_bias",
        "logprobs",
        "max_tokens",
        "n",
        "presence_penalty",
        "seed",
        "stop",
        "stream",
        "suffix",
        "temperature",
        "top_p",
        "user",
    }

    progress_prompt: str = "Generating..."

    def __init__(
        self,
        cache_file: str | None = None,
        max_retries: int | None = None,
        base_url: str | None = None,
        api_key: str | None = None,
        max_qps: Optional[float] = None,
        max_qpm: float = 100,
    ):
        """
        Initialize the client.

        Args:
            - cache_file: the file to cache the results.
            - max_retries: the maximum number of retries to make, defaults to `None`.
            - base_url: OpenAI Compatible API base URL.
            - api_key: OpenAI API key.
            - max_qps: maximum queries per second (rate limit), defaults to `None` (falls back to max_qpm).
            - max_qpm: maximum queries per minute (rate limit), defaults to `100`.
        """

        super().__init__(
            cache_file=cache_file, 
            max_retries=max_retries, 
            max_qps=max_qps,
            max_qpm=max_qpm
        )

        # initialize the client
        self.client = AsyncOpenAI(
            base_url=base_url or os.getenv("OPENAI_API_BASE"),
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
        )

    async def make_request_async(self, request: Dict[str, Any], **kwargs) -> Dict | None:
        """
        Make a single request to the OpenAI API and cache the response.

        Args:
            request: the request to make
            **kwargs: Additional arguments that will be passed to the OpenAI API call if they are in SUPPORT_ARGS

        Returns:
            bool: True if the request was successful and cached, False if it failed
        """

        try:
            assert "prompt" in request, "`prompt` is required in the request"
            assert "model" in request, "`model` is required in the request"

            gen_kwargs = {}
            for key in self.SUPPORT_ARGS:
                if key in request:
                    gen_kwargs[key] = request[key]
                else:
                    if key in kwargs and kwargs[key] is not None:
                        gen_kwargs[key] = kwargs[key]
            # add timeout if it is specified
            if "timeout" in kwargs:
                gen_kwargs["timeout"] = kwargs["timeout"]
            # request for the response
            response: Completion = await self.client.completions.create(**gen_kwargs)
            response_dict = response.model_dump()
            return response_dict

        except Exception as e:
            print(f"{type(e).__name__}: {e}")
            return None

    def collect_responses(self, requests: List[Dict]) -> List[Completion]:
        """
        Get the cached responses for a list of requests and convert them to `Completion` objects.

        Args:
            requests: list of requests to get responses for
        Returns:
            list of Completion objects containing the cached responses
        """

        # First get the cached responses as dictionaries
        cached_responses = [self.cache.collect_result(request) for request in requests]

        # Then convert dictionaries to Completion objects
        completions = []
        for cached_response in cached_responses:
            if cached_response is None:
                completion = None
            else:
                completion = Completion.model_validate(cached_response)
            completions.append(completion)

        return completions



import os
import pytest
from unittest.mock import AsyncMock, MagicMock, patch

from fluxllm.clients import FluxOpenAICompletion


@pytest.fixture
def mock_openai_completion():
    """Mock the OpenAI Completion API response"""
    mock_response = MagicMock()
    mock_response.model_dump.return_value = {
        "id": "cmpl-123",
        "object": "text_completion",
        "created": 1677858242,
        "model": "text-davinci-003",
        "choices": [
            {
                "text": "This is a test completion response.",
                "index": 0,
                "logprobs": None,
                "finish_reason": "stop"
            }
        ],
        "usage": {
            "prompt_tokens": 5,
            "completion_tokens": 7,
            "total_tokens": 12
        }
    }
    return mock_response


@pytest.mark.asyncio
async def test_flux_openai_completion(mock_openai_completion):
    """Test the FluxOpenAICompletion client"""
    
    # Create a temporary cache file
    cache_file = "test_cache.jsonl"
    if os.path.exists(cache_file):
        os.remove(cache_file)
    
    # Mock the AsyncOpenAI client
    with patch("fluxllm.clients.openai.AsyncOpenAI") as mock_async_openai:
        # Set up the mock
        mock_client = MagicMock()
        mock_completions = MagicMock()
        mock_completions.create = AsyncMock(return_value=mock_openai_completion)
        mock_client.completions = mock_completions
        mock_async_openai.return_value = mock_client
        
        # Create the client with a dummy API key
        client = FluxOpenAICompletion(cache_file=cache_file, api_key="dummy_key")
        
        # Create a test request
        request = {
            "model": "text-davinci-003",
            "prompt": "Complete this sentence:"
        }
        
        # Make the request
        response = await client.make_request_async(request)
        
        # Check that the request was made with the correct parameters
        mock_client.completions.create.assert_called_once_with(
            model="text-davinci-003",
            prompt="Complete this sentence:"
        )
        
        # Check that the response was returned correctly
        assert response["choices"][0]["text"] == "This is a test completion response."
    
    # Clean up
    if os.path.exists(cache_file):
        os.remove(cache_file)


from fluxllm.clients import FluxOpenAIChat


def test_basic():

    client = FluxOpenAIChat(
        base_url="http://127.0.0.1:8008/v1",
        api_key="test",
        cache_file="cache.jsonl",
        max_retries=3,
        max_qps=5,  # Add rate limiting of 5 requests per second
    )

    request = {
        "messages": [{
            "role": "user",
            "content": "Hello, world!"
        },],
        "model": "gpt-4o",
        "max_tokens": 100,
        "temperature": 0.5,
        "top_p": 1,
    }
    # requests is a list of requests
    requests = [{**request, "id": f"id-{i}"} for i in range(10)]

    responses = client.request(
        requests=requests,
        model="gpt-4o",
        max_tokens=100,
        temperature=0.5,
        top_p=1.0,
    )

    assert len(responses) == len(requests)


def test_qpm_rate_limiting():
    """Test rate limiting using QPM instead of QPS"""
    
    client = FluxOpenAIChat(
        base_url="http://127.0.0.1:8008/v1",
        api_key="test",
        cache_file="cache.jsonl",
        max_retries=3,
        max_qps=None,  # Disable QPS rate limiting
        max_qpm=60,    # Set QPM rate limiting to 60 requests per minute
    )

    request = {
        "messages": [{
            "role": "user",
            "content": "Hello, world!"
        },],
        "model": "gpt-4o",
        "max_tokens": 100,
        "temperature": 0.5,
        "top_p": 1,
    }
    # requests is a list of requests
    requests = [{**request, "id": f"id-{i}"} for i in range(5)]

    responses = client.request(
        requests=requests,
        model="gpt-4o",
        max_tokens=100,
        temperature=0.5,
        top_p=1.0,
    )

    assert len(responses) == len(requests)


def test_dual_rate_limiting():
    """Test rate limiting using both QPS and QPM simultaneously"""
    
    client = FluxOpenAIChat(
        base_url="http://127.0.0.1:8008/v1",
        api_key="test",
        cache_file="cache.jsonl",
        max_retries=3,
        max_qps=10,    # 10 requests per second
        max_qpm=120,   # 120 requests per minute (2 per second)
    )

    request = {
        "messages": [{
            "role": "user",
            "content": "Hello, world!"
        },],
        "model": "gpt-4o",
        "max_tokens": 100,
        "temperature": 0.5,
        "top_p": 1,
    }
    # requests is a list of requests
    requests = [{**request, "id": f"id-{i}"} for i in range(15)]

    responses = client.request(
        requests=requests,
        model="gpt-4o",
        max_tokens=100,
        temperature=0.5,
        top_p=1.0,
    )

    assert len(responses) == len(requests)
    # Both rate limits should be enforced, with the more restrictive one (QPM in this case)
    # determining the actual concurrency



from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    TimeRemainingColumn,
    TimeElapsedColumn,
)


def create_progress():
    """
    Create a rich progress bar.
    """
    return Progress(
        TextColumn("[bold blue]{task.description}"),
        BarColumn(),
        TaskProgressColumn(show_speed=True),
        TextColumn("[green]{task.completed}/{task.total}"),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    )



import asyncio
import os
import random
from typing import Any, Dict, List, Optional

from aiolimiter import AsyncLimiter
from openai.types.chat import ChatCompletion

from fluxllm.client_utils import FluxCache, create_progress


class BaseClient:
    """
    A client that allows for concurrent requests to the same API.
    
    Rate limiting is implemented using both queries-per-second (QPS) and 
    queries-per-minute (QPM) controls. If both are specified, both limits 
    will be enforced simultaneously. Concurrency is automatically calculated 
    based on the more restrictive of the two rate limits.
    """

    def __init__(
        self,
        cache_file: str | None = None,
        max_retries: int | None = None,
        max_qps: Optional[float] = None,  # None means no QPS rate limiting
        max_qpm: float = 100,  # Default to 100 requests per minute
        progress_msg: str = "Requesting...",
    ):
        """
        Initialize the client.

        Args:
            cache_file: the cache file to use.
            max_retries: the maximum number of retries to make, defaults to `None`.
            max_qps: maximum queries per second (rate limit), defaults to `None`.
            max_qpm: maximum queries per minute (rate limit), defaults to `100`.
            progress_msg: the message to display in the progress bar, defaults to `Requesting...`.
        """
        if cache_file is None:
            cache_file = os.getenv("CACHE_FILE", "cache.jsonl")

        # initialize the cache client
        self.cache = FluxCache(cache_file)
        self.lock = asyncio.Lock()
        self.max_retries = max_retries
        self.max_qps = max_qps
        self.max_qpm = max_qpm
        
        # Configure rate limiters for both QPS and QPM if specified
        self.qps_limiter = None
        self.qpm_limiter = None
        
        if max_qps is not None and max_qps > 0:
            self.qps_limiter = AsyncLimiter(max_rate=max_qps, time_period=1.0)
            
        if max_qpm > 0:
            self.qpm_limiter = AsyncLimiter(max_rate=max_qpm, time_period=60.0)
            
        # Calculate optimal concurrency based on rate limits
        if max_qps is not None and max_qps > 0:
            qps_concurrency = max(1, int(max_qps))
        else:
            qps_concurrency = float('inf')
            
        if max_qpm > 0:
            qpm_concurrency = max(1, int(max_qpm / 60))
        else:
            qpm_concurrency = float('inf')
            
        # Use the more restrictive concurrency limit
        self.concurrency = min(qps_concurrency, qpm_concurrency)
        if self.concurrency == float('inf'):
            self.concurrency = 1  # Default if no rate limits are set
            
        self.progress_msg = progress_msg

    async def save_to_cache_thread_safe(self, sample: Dict, response: Dict, save_request: bool = False):
        async with self.lock:
            self.cache.save_to_cache(sample, response, save_request=save_request)

    async def execute_with_rate_limiting(self, request: Dict, **kwargs):
        """
        Execute a request with rate limiting applied.
        
        This method applies both QPS and QPM rate limiting if configured.
        
        Args:
            request: The request to execute
            **kwargs: Additional arguments to pass to make_request_async
            
        Returns:
            The response from make_request_async
        """
        # Apply QPS rate limiting if configured
        if self.qps_limiter is not None:
            async with self.qps_limiter:
                # Apply QPM rate limiting if configured
                if self.qpm_limiter is not None:
                    async with self.qpm_limiter:
                        return await self.make_request_async(request, **kwargs)
                else:
                    return await self.make_request_async(request, **kwargs)
        # Only apply QPM rate limiting if QPS is not configured
        elif self.qpm_limiter is not None:
            async with self.qpm_limiter:
                return await self.make_request_async(request, **kwargs)
        # No rate limiting
        else:
            return await self.make_request_async(request, **kwargs)

    async def request_async(
        self,
        requests: List[Dict[str, Any]],
        save_request: bool = False,
        **kwargs,
    ) -> None:
        """
        Make requests for all uncached samples in given list.
        This function is designed to handle the generation of multiple responses in a batch.
        It uses rate limiters to control the request rate.

        Args:
            requests: list of requests to generate responses for
            save_request: whether to save the request in the cache
            **kwargs: additional arguments to pass to request_async
        """

        # create a queue to hold the samples
        task_queue = asyncio.Queue()
        for request in requests:
            await task_queue.put(request)
        failure_counts = {self.cache.hash(request): 0 for request in requests}

        async def after_failure(request: Dict):
            failure_counts[self.cache.hash(request)] += 1
            if self.max_retries is None:
                print(f"Re-queue failed request: {self.cache.hash(request)}, failed {failure_counts[self.cache.hash(request)]} times.", flush=True)
                await task_queue.put(request)
                await asyncio.sleep(random.randint(3, 10))
            else:
                if failure_counts[self.cache.hash(request)] < self.max_retries:
                    print(f"Re-queue failed request: {self.cache.hash(request)}, failed {failure_counts[self.cache.hash(request)]} times.", flush=True)
                    await task_queue.put(request)
                    await asyncio.sleep(random.randint(3, 10))
                else:
                    print(f"Request failed after {self.max_retries} retries. Aborting this request.", flush=True)
                    progress.advance(task)

        async def worker():
            while not task_queue.empty():
                request = await task_queue.get()
                response = await self.execute_with_rate_limiting(request, **kwargs)
                
                if response is not None:
                    await self.save_to_cache_thread_safe(request, response, save_request=save_request)
                    progress.advance(task)
                    print(f"Request succeeded for request: {self.cache.hash(request)}", flush=True)
                else:
                    await after_failure(request)
                task_queue.task_done()

        with create_progress() as progress:
            task = progress.add_task(f"[cyan]{self.progress_msg}", total=len(requests))
            workers = [asyncio.create_task(worker()) for _ in range(self.concurrency)]

            await task_queue.join()
            for worker_task in workers:
                worker_task.cancel()

    def request(self, requests: List[Dict[str, Any]], save_request: bool = False, **kwargs) -> List[ChatCompletion | None]:
        """
        Make requests for all uncached samples in given list.
        This is a synchronous wrapper around request_async.

        Args:
            requests: List of requests to make
            **kwargs: additional arguments to pass to request_async
        """

        # get the samples that are not cached
        remaining_requests = [request for request in requests if not self.cache.is_cached(request)]
        print(f"Remaining {len(remaining_requests)} requests to generate", flush=True)

        # request the responses
        asyncio.run(self.request_async(requests=remaining_requests, save_request=save_request, **kwargs))

        # collect the responses
        responses = self.collect_responses(requests)

        return responses

    async def make_request_async(self, request: Dict, **kwargs) -> Dict | None:
        """
        Make a single request to the API and cache the response.
        """
        raise NotImplementedError("Implement this in the subclass")

    def collect_responses(self, requests: List[Dict]) -> List[ChatCompletion | None]:
        """
        Collect the responses from the cache.
        """
        raise NotImplementedError("Implement this in the subclass")
